"""Tests for the svg2pdf preprocessor"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from nbformat import v4 as nbformat
from unittest.mock import patch, Mock

from .base import PreprocessorTestsBase
from ..svg2pdf import SVG2PDFPreprocessor
from ...tests.utils import onlyif_cmds_exist


class Testsvg2pdf(PreprocessorTestsBase):
    """Contains test functions for svg2pdf.py"""

    simple_svg = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<!-- Created with Inkscape (http://www.inkscape.org/) -->
<svg
   xmlns:svg="http://www.w3.org/2000/svg"
   xmlns="http://www.w3.org/2000/svg"
   version="1.0"
   x="0.00000000"
   y="0.00000000"
   width="500.00000"
   height="500.00000"
   id="svg2">
  <defs
     id="defs4" />
  <g
     id="layer1">
    <rect
       width="300.00000"
       height="300.00000"
       x="100.00000"
       y="100.00000"
       style="opacity:1.0000000;fill:none;fill-opacity:1.0000000;fill-rule:evenodd;stroke:#000000;stroke-width:8.0000000;stroke-linecap:round;stroke-linejoin:round;stroke-miterlimit:4.0000000;stroke-dasharray:none;stroke-dashoffset:0.00000000;stroke-opacity:1.0000000"
       id="rect5719" />
  </g>
</svg>"""

    def build_notebook(self):
        """Build a reveal slides notebook in memory for use with tests.
        Overrides base in PreprocessorTestsBase"""

        outputs = [nbformat.new_output(output_type='display_data',
                                       data={'image/svg+xml':self.simple_svg})
                  ]

        cells=[nbformat.new_code_cell(source="", execution_count=1, outputs=outputs)]

        return nbformat.new_notebook(cells=cells)


    def build_preprocessor(self, **kwargs):
        """Make an instance of a preprocessor"""
        preprocessor = SVG2PDFPreprocessor(**kwargs)
        preprocessor.enabled = True
        return preprocessor


    def test_constructor(self):
        """Can a SVG2PDFPreprocessor be constructed?"""
        self.build_preprocessor()


    @onlyif_cmds_exist('inkscape')
    def test_output(self):
        """Test the output of the SVG2PDFPreprocessor"""
        nb = self.build_notebook()
        res = self.build_resources()
        preprocessor = self.build_preprocessor()
        nb, res = preprocessor(nb, res)
        self.assertIn('application/pdf', nb.cells[0].outputs[0].data)

    @patch('subprocess.Popen')
    def test_inkscape_version_default(self, mock_popen):
        mock_popen().communicate.return_value = (b'Inkscape 0.92.3 (2405546, 2018-03-11)', b'')
        mock_popen().returncode = 0

        preprocessor = self.build_preprocessor()
        assert preprocessor.inkscape_version == '0.92.3'

    def test_inkscape_pre_v1_command(self):
        preprocessor = self.build_preprocessor(inkscape_version='0.92.3')
        assert preprocessor.command == '0.92.3'

    def test_inkscape_pre_v1_command(self):
        preprocessor = self.build_preprocessor(inkscape='fake-inkscape', inkscape_version='0.92.3')
        assert preprocessor.command == 'fake-inkscape --without-gui --export-pdf="{to_filename}" "{from_filename}"'

    def test_inkscape_v1_command(self):
        preprocessor = self.build_preprocessor(inkscape='fake-inkscape', inkscape_version='1.0beta2')
        assert preprocessor.command == 'fake-inkscape --export-filename="{to_filename}" "{from_filename}"'



"""PostProcessor for serving reveal.js HTML slideshows."""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function

import os
import webbrowser
import threading

from tornado import web, ioloop, httpserver, log, gen
from tornado.httpclient import AsyncHTTPClient
from traitlets import Bool, Unicode, Int

from .base import PostProcessorBase


class ProxyHandler(web.RequestHandler):
    """handler the proxies requests from a local prefix to a CDN"""
    @gen.coroutine
    def get(self, prefix, url):
        """proxy a request to a CDN"""
        proxy_url = "/".join([self.settings['cdn'], url])
        client = self.settings['client']
        response = yield client.fetch(proxy_url)

        for header in ["Content-Type", "Cache-Control", "Date", "Last-Modified", "Expires"]:
            if header in response.headers:
                self.set_header(header, response.headers[header])
        self.finish(response.body)


class ServePostProcessor(PostProcessorBase):
    """Post processor designed to serve files
    
    Proxies reveal.js requests to a CDN if no local reveal.js is present
    """


    open_in_browser = Bool(True,
        help="""Should the browser be opened automatically?"""
    ).tag(config=True)

    browser = Unicode(u'', 
                      help="""Specify what browser should be used to open slides. See
                      https://docs.python.org/3/library/webbrowser.html#webbrowser.register
                      to see how keys are mapped to browser executables. If 
                      not specified, the default browser will be determined 
                      by the `webbrowser` 
                      standard library module, which allows setting of the BROWSER 
                      environment variable to override it.
                      """).tag(config=True)

    reveal_cdn = Unicode("https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.5.0",
                         help="""URL for reveal.js CDN.""").tag(config=True)
    reveal_prefix = Unicode("reveal.js",
                            help="URL prefix for reveal.js").tag(config=True)
    ip = Unicode("127.0.0.1",
                 help="The IP address to listen on.").tag(config=True)
    port = Int(8000, help="port for the server to listen on.").tag(config=True)

    def postprocess(self, input):
        """Serve the build directory with a webserver."""
        dirname, filename = os.path.split(input)
        handlers = [
            (r"/(.+)", web.StaticFileHandler, {'path' : dirname}),
            (r"/", web.RedirectHandler, {"url": "/%s" % filename})
        ]
        
        if ('://' in self.reveal_prefix or self.reveal_prefix.startswith("//")):
            # reveal specifically from CDN, nothing to do
            pass
        elif os.path.isdir(os.path.join(dirname, self.reveal_prefix)):
            # reveal prefix exists
            self.log.info("Serving local %s", self.reveal_prefix)
        else:
            self.log.info("Redirecting %s requests to %s", self.reveal_prefix, self.reveal_cdn)
            handlers.insert(0, (r"/(%s)/(.*)" % self.reveal_prefix, ProxyHandler))
        
        app = web.Application(handlers,
                              cdn=self.reveal_cdn,
                              client=AsyncHTTPClient(),
                              )
        
        # hook up tornado logging to our logger
        log.app_log = self.log

        http_server = httpserver.HTTPServer(app)
        http_server.listen(self.port, address=self.ip)
        url = "http://%s:%i/%s" % (self.ip, self.port, filename)
        print("Serving your slides at %s" % url)
        print("Use Control-C to stop this server")
        if self.open_in_browser:
            try:
                browser = webbrowser.get(self.browser or None)
                b = lambda: browser.open(url, new=2)
                threading.Thread(target=b).start()
            except webbrowser.Error as e:
                self.log.warning('No web browser found: %s.' % e)
                browser = None

        try:
            ioloop.IOLoop.instance().start()
        except KeyboardInterrupt:
            print("\nInterrupted")

def main(path):
    """allow running this module to serve the slides"""
    server = ServePostProcessor()
    server(path)
    
if __name__ == '__main__':
    import sys
    main(sys.argv[1])



"""Tests for RSTExporter"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import io
import pytest
import re

import nbformat
from nbformat import v4

from .base import ExportersTestsBase
from ..rst import RSTExporter
from ...tests.utils import onlyif_cmds_exist


class TestRSTExporter(ExportersTestsBase):
    """Tests for RSTExporter"""

    exporter_class = RSTExporter
    should_include_raw = ['rst']

    def test_constructor(self):
        """
        Can a RSTExporter be constructed?
        """
        RSTExporter()


    @onlyif_cmds_exist('pandoc')
    def test_export(self):
        """
        Can a RSTExporter export something?
        """
        (output, resources) = RSTExporter().from_filename(self._get_notebook())
        assert len(output) > 0

    @onlyif_cmds_exist('pandoc')
    def test_export_nbformat_template_v5_name(self):
        """
        We support the nbconvert v5 filename, but with a deprecation warning.
        """
        with pytest.warns(DeprecationWarning):
            (output, resources) = RSTExporter(template_file='rst.tpl').from_filename(self._get_notebook())
        assert len(output) > 0

    @onlyif_cmds_exist('pandoc')
    def test_empty_code_cell(self):
        """No empty code cells in rst"""
        nbname = self._get_notebook()
        with io.open(nbname, encoding='utf8') as f:
            nb = nbformat.read(f, 4)

        exporter = self.exporter_class()

        (output, resources) = exporter.from_notebook_node(nb)
        # add an empty code cell
        nb.cells.append(
            v4.new_code_cell(source="")
        )
        (output2, resources) = exporter.from_notebook_node(nb)
        # adding an empty code cell shouldn't change output
        self.assertEqual(output.strip(), output2.strip())

    @onlyif_cmds_exist('pandoc')
    def test_png_metadata(self):
        """
        Does RSTExporter treat pngs with width/height metadata correctly?
        """
        (output, resources) = RSTExporter().from_filename(
            self._get_notebook(nb_name="pngmetadata.ipynb"))
        assert len(output) > 0
        check_for_png = re.compile(
            r'.. image::.*?\n\s+(.*?)\n\s*\n',
            re.DOTALL)
        result = check_for_png.search(output)
        assert result is not None
        attr_string = result.group(1)
        assert ':width:' in attr_string
        assert ':height:' in attr_string
        assert 'px' in attr_string



# -*- coding: utf-8 -*-
"""HTML Exporter class"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import mimetypes
import base64

from traitlets import default, Unicode, Bool
from traitlets.config import Config
from jupyter_core.paths import jupyter_path
from jinja2 import contextfilter
from jinja2.loaders import split_template_path
import jinja2

from nbconvert.filters.highlight import Highlight2HTML
from nbconvert.filters.markdown_mistune import IPythonRenderer, MarkdownWithMath

from .templateexporter import TemplateExporter


class HTMLExporter(TemplateExporter):
    """
    Exports a basic HTML document.  This exporter assists with the export of
    HTML.  Inherit from it if you are writing your own HTML template and need
    custom preprocessors/filters.  If you don't need custom preprocessors/
    filters, just change the 'template_file' config option.
    """
    export_from_notebook = "HTML"

    anchor_link_text = Unicode(u'¶',
        help="The text used as the text for anchor links.").tag(config=True)

    exclude_anchor_links = Bool(False,
        help="If anchor links should be included or not.").tag(config=True)

    require_js_url = Unicode(
        "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.1.10/require.min.js",
        help="""
        URL to load require.js from.

        Defaults to loading from cdnjs.
        """
    ).tag(config=True)

    jquery_url = Unicode(
        "https://cdnjs.cloudflare.com/ajax/libs/jquery/2.0.3/jquery.min.js",
        help="""
        URL to load jQuery from.

        Defaults to loading from cdnjs.
        """
    ).tag(config=True)

    @default('file_extension')
    def _file_extension_default(self):
        return '.html'

    @default('template_name')
    def _template_name_default(self):
        return 'lab'

    @default('template_data_paths')
    def _template_data_paths_default(self):
        return jupyter_path("nbconvert", "templates", "html")


    theme = Unicode('light',
                    help='Template specific theme(e.g. the JupyterLab CSS theme for the lab template)'
    ).tag(config=True)

    output_mimetype = 'text/html'

    @property
    def default_config(self):
        c = Config({
            'NbConvertBase': {
                'display_data_priority' : ['application/vnd.jupyter.widget-state+json',
                                           'application/vnd.jupyter.widget-view+json',
                                           'application/javascript',
                                           'text/html',
                                           'text/markdown',
                                           'image/svg+xml',
                                           'text/latex',
                                           'image/png',
                                           'image/jpeg',
                                           'text/plain'
                                          ]
                },
            'CSSHTMLHeaderPreprocessor':{
                'enabled':True
                },
            'HighlightMagicsPreprocessor': {
                'enabled':True
                }
            })
        c.merge(super().default_config)
        return c

    @contextfilter
    def markdown2html(self, context, source):
        """Markdown to HTML filter respecting the anchor_link_text setting"""
        cell = context.get('cell', {})
        attachments = cell.get('attachments', {})
        renderer = IPythonRenderer(escape=False, attachments=attachments,
                                   anchor_link_text=self.anchor_link_text,
                                   exclude_anchor_links=self.exclude_anchor_links)
        return MarkdownWithMath(renderer=renderer).render(source)

    def default_filters(self):
        for pair in super().default_filters():
            yield pair
        yield ('markdown2html', self.markdown2html)

    def from_notebook_node(self, nb, resources=None, **kw):
        langinfo = nb.metadata.get('language_info', {})
        lexer = langinfo.get('pygments_lexer', langinfo.get('name', None))
        highlight_code = self.filters.get('highlight_code', Highlight2HTML(pygments_lexer=lexer, parent=self))
        self.register_filter('highlight_code', highlight_code)
        return super().from_notebook_node(nb, resources, **kw)

    def _init_resources(self, resources):
        def resources_include_css(name):
            env = self.environment
            code = """<style type="text/css">\n%s</style>""" % (env.loader.get_source(env, name)[0])
            return jinja2.Markup(code)

        def resources_include_js(name):
            env = self.environment
            code = """<script>\n%s</script>""" % (env.loader.get_source(env, name)[0])
            return jinja2.Markup(code)

        def resources_include_url(name):
            env = self.environment
            mime_type, encoding = mimetypes.guess_type(name)
            try:
                # we try to load via the jinja loader, but that tries to load
                # as (encoded) text
                data = env.loader.get_source(env, name)[0].encode('utf8')
            except UnicodeDecodeError:
                # if that fails (for instance a binary file, png or ttf)
                # we mimic jinja2
                pieces = split_template_path(name)
                searchpaths = self.get_template_paths()
                for searchpath in searchpaths:
                    filename = os.path.join(searchpath, *pieces)
                    print(filename, os.path.exists(filename))
                    if os.path.exists(filename):
                        with open(filename, "rb") as f:
                            data = f.read()
                            break
                else:
                    raise ValueError("No file %r found in %r" % (name, searchpaths))
            data = base64.b64encode(data)
            data = data.replace(b'\n', b'').decode('ascii')
            src = 'data:{mime_type};base64,{data}'.format(mime_type=mime_type, data=data)
            return jinja2.Markup(src)
        resources = super()._init_resources(resources)
        resources['theme'] = self.theme
        resources['include_css'] = resources_include_css
        resources['include_js'] = resources_include_js
        resources['include_url'] = resources_include_url
        resources['require_js_url'] = self.require_js_url
        resources['jquery_url'] = self.jquery_url
        return resources



# coding: utf-8
"""Tests for conversions from markdown to other formats"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import re
from copy import copy
from functools import partial
from html import unescape

from ...tests.utils import onlyif_cmds_exist
from ...tests.base import TestsBase
from ..pandoc import convert_pandoc
from ..markdown import markdown2html

from jinja2 import Environment


class TestMarkdown(TestsBase):

    tests = [
        '*test',
        '**test',
        '*test*',
        '_test_',
        '__test__',
        '__*test*__',
        '**test**',
        '#test',
        '##test',
        'test\n----',
        'test [link](https://google.com/)',
    ]

    tokens = [
        '*test',
        '**test',
        'test',
        'test',
        'test',
        'test',
        'test',
        'test',
        'test',
        'test',
        ('test', 'https://google.com/'),
    ]

    @onlyif_cmds_exist('pandoc')
    def test_markdown2latex(self):
        """markdown2latex test"""
        for index, test in enumerate(self.tests):
            self._try_markdown(
                partial(
                    convert_pandoc, from_format='markdown', to_format='latex'),
                test, self.tokens[index])

    @onlyif_cmds_exist('pandoc')
    def test_markdown2latex_markup(self):
        """markdown2latex with markup kwarg test"""
        # This string should be passed through unaltered with pandoc's
        # markdown_strict reader
        s = '1) arabic number with parenthesis'
        self.assertEqual(convert_pandoc(s, 'markdown_strict', 'latex'), s)
        # This string should be passed through unaltered with pandoc's
        # markdown_strict+tex_math_dollars reader
        s = r'$\alpha$ latex math'
        # sometimes pandoc uses $math$, sometimes it uses \(math\)
        expected = re.compile(r'(\$|\\\()\\alpha(\$|\\\)) latex math')

        assertRegex = self.assertRegex

        assertRegex(
            convert_pandoc(s, 'markdown_strict+tex_math_dollars', 'latex'),
            expected)

    @onlyif_cmds_exist('pandoc')
    def test_pandoc_extra_args(self):
        # pass --no-wrap
        s = '\n'.join([
            "#latex {{long_line | md2l(['--wrap=none'])}}",
            "#rst {{long_line | md2r(['--columns', '5'])}}",
        ])
        long_line = ' '.join(['long'] * 30)
        env = Environment()
        env.filters.update({
            'md2l': lambda code, extra_args: convert_pandoc(
                code, from_format='markdown', to_format='latex',
                extra_args=extra_args),
            'md2r': lambda code, extra_args: convert_pandoc(
                code, from_format='markdown', to_format='rst',
                extra_args=extra_args),
        })
        tpl = env.from_string(s)
        rendered = tpl.render(long_line=long_line)
        _, latex, rst = rendered.split('#')

        self.assertEqual(latex.strip(), 'latex %s' % long_line)
        self.assertEqual(rst.strip(), 'rst %s' % long_line.replace(' ', '\n'))

    def test_markdown2html(self):
        """markdown2html test"""
        for index, test in enumerate(self.tests):
            self._try_markdown(markdown2html, test, self.tokens[index])

    def test_markdown2html_heading_anchors(self):
        for md, tokens in [('# test', ('<h1', '>test', 'id="test"',
                                       u'&#182;</a>', "anchor-link")),
                           ('###test head space',
                            ('<h3', '>test head space', 'id="test-head-space"',
                             u'&#182;</a>', "anchor-link"))]:
            self._try_markdown(markdown2html, md, tokens)

    def test_markdown2html_math(self):
        # Mathematical expressions not containing <, >, &
        # should be passed through unaltered
        # all the "<", ">", "&" must be escaped correctly
        cases = [(
            "\\begin{equation*}\n" +
            ("\\left( \\sum_{k=1}^n a_k b_k \\right)^2 "
             "\\leq \\left( \\sum_{k=1}^n a_k^2 \\right) "
             "\\left( \\sum_{k=1}^n b_k^2 \\right)\n") +
            "\\end{equation*}"),
            ("$$\n"
             "a = 1 *3* 5\n"
             "$$"),
            "$ a = 1 *3* 5 $",
            "$s_i = s_{i}\n$",
            "$a<b&b<lt$",
            "$a<b&lt;b>a;a-b<0$",
            "$<k'>$",
            "$$a<b&b<lt$$",
            "$$a<b&lt;b>a;a-b<0$$",
            "$$<k'>$$",
            ("$$x\n"
             "=\n"
             "2$$"),
            ("$$\n"
             "b =  \\left[\n"
             "P\\left(\\right)\n"
             "- (l_1\\leftrightarrow l_2\n)"
             "\\right]\n"
             "$$"),
            ("\\begin{equation*}\n"
             "x = 2 *55* 7\n"
             "\\end{equation*}"),
            """$
\\begin{tabular}{ l c r }
  1 & 2 & 3 \\
  4 & 5 & 6 \\
  7 & 8 & 9 \\
\\end{tabular}$"""]

        for case in cases:
            result = markdown2html(case)
            # find the equation in the generated texts
            search_result = re.search(r"\$.*\$", result, re.DOTALL)
            if search_result is None:
                search_result = re.search(
                    "\\\\begin\\{equation.*\\}.*\\\\end\\{equation.*\\}",
                    result, re.DOTALL)
            math = search_result.group(0)
            # the resulting math part can not contain "<", ">" or
            # "&" not followed by "lt;", "gt;", or "amp;".
            self.assertNotIn("<", math)
            self.assertNotIn(">", math)
            self.assertNotRegex(math, "&(?![gt;|lt;|amp;])")
            # the result should be able to be unescaped correctly
            self.assertEqual(case, unescape(math))

    def test_markdown2html_math_mixed(self):
        """ensure markdown between inline and inline-block math works and
        test multiple LaTeX markup syntaxes.
        """
        case = """The entries of \\\\(C\\\\) are given by the exact formula:
$$
C_{ik} = \\sum_{j=1}^n A_{ij} B_{jk},
$$
but you can _implement_ this computation in many ways.
$\approx 2mnp$ flops are needed for \\\\[ C_{ik} = \\sum_{j=1}^n A_{ij} B_{jk} \\\\].
Also check empty math blocks work correctly:
$$$$
\\\\[\\\\]"""
        output_check = (case.replace("_implement_", "<em>implement</em>")
                            .replace("\\\\(", "$").replace("\\\\)", "$")
                            .replace("\\\\[", "$$").replace("\\\\]", "$$"))
        # these replacements are needed because we use $ and $$ in our html output
        self._try_markdown(markdown2html, case, output_check)

    def test_markdown2html_math_paragraph(self):
        """these should all parse without modification"""
        cases = [
            # https://github.com/ipython/ipython/issues/6724
            """Water that is stored in $t$, $s_t$, must equal the storage content of the previous stage,
$s_{t-1}$, plus a stochastic inflow, $I_t$, minus what is being released in $t$, $r_t$.
With $s_0$ defined as the initial storage content in $t=1$, we have""",
            # https://github.com/jupyter/nbviewer/issues/420
            """$C_{ik}$
$$
C_{ik} = \\sum_{j=1}
$$
$C_{ik}$""",
            """$m$
$$
C = \begin{pmatrix}
          0 & 0 & 0 & \\cdots & 0 & 0 & -c_0 \\
          0 & 0 & 0 & \\cdots & 0 & 1 & -c_{m-1}
    \\end{pmatrix}
$$
$x^m$""",
            """$r=\\overline{1,n}$
$$ {\bf
b}_{i}^{r}(t)=(1-t)\\,{\bf b}_{i}^{r-1}(t)+t\\,{\bf b}_{i+1}^{r-1}(t),\\:
 i=\\overline{0,n-r}, $$
i.e. the $i^{th}$"""
        ]

        for case in cases:
            s = markdown2html(case)
            self.assertIn(case, unescape(s))

    @onlyif_cmds_exist('pandoc')
    def test_markdown2rst(self):
        """markdown2rst test"""

        #Modify token array for rst, escape asterisk
        tokens = copy(self.tokens)
        tokens[0] = r'\*test'
        tokens[1] = r'\**test'

        for index, test in enumerate(self.tests):
            self._try_markdown(
                partial(
                    convert_pandoc, from_format='markdown', to_format='rst'),
                test, tokens[index])

    def _try_markdown(self, method, test, tokens):
        results = method(test)
        if isinstance(tokens, (str,)):
            self.assertIn(tokens, results)
        else:
            for token in tokens:
                self.assertIn(token, results)
